{
open Symbol
open Fc_ast_parse
open Fc_ast_type

(*
 * File position.
 *)
let current_line = ref 1
let current_schar = ref 0

(*
 * Advance a line.
 *)
let set_next_line lexbuf =
   incr current_line;
   current_schar := Lexing.lexeme_end lexbuf

(*
 * Get the position of the current lexeme.
 * We assume it is all on one line.
 *)
let set_lexeme_position lexbuf =
   let line = !current_line in
   let schar = Lexing.lexeme_start lexbuf - !current_schar in
   let echar = Lexing.lexeme_end lexbuf - !current_schar in
   let file = Fc_ast_state.current_file () in
   let pos = file, line, schar, line, echar in
      Fc_ast_state.set_current_position pos;
      pos

(*
 * TODO: put your other ML utilities here.
 *)
}

(*
 * Here are some example regular expressions.
 *)
let white = [' ' '\t']+

(*
 * An identifier.
 *)
let name_prefix = ['_' 'A'-'Z' 'a'-'z']
let name_suffix = ['_' 'A'-'Z' 'a'-'z' '0'-'9']
let name = name_prefix name_suffix *

(*
 * Numbers.
 *)
let decimal = ['0'-'9']+

(*
 * Some special symbols.
 *)
let special = [';' '+' '-' ')' '(']

(*
 * Main lexer.
 *)
rule main = parse
   (*
    * Handle white space.
    *)
   white
      { (* Ignore white space *)
        main lexbuf
      }
 | '\n'
      { (* Keep track of the current line number *)
        set_next_line lexbuf;
        main lexbuf
      }

   (*
    * Decimal number.
    * You also need to define octal, hex, and float constants.
    *)
 | decimal
      { (* This is a decimal number *)
        let pos = set_lexeme_position lexbuf in
           TokInt (int_of_string (Lexing.lexeme lexbuf), pos)
      }

   (*
    * Identifiers, including keywords.
    *)
 | name
      { let pos = set_lexeme_position lexbuf in
        let id = Lexing.lexeme lexbuf in
           (* Some example keywords *)
           match id with
              "typedef" -> TokTypedef pos
            | "if" -> TokIf pos
            | "else" -> TokElse pos
            | "for" -> TokFor pos
            | "while" -> TokWhile pos
            | "return" -> TokReturn pos
            | "break" -> TokBreak pos
            | "struct" -> TokStruct pos
            | id ->
                 (*
                  * If it is not a keyword, then it is either a variable
                  * or a type name.  Convert the string to a symbol.
                  *)
                 let sym = Symbol.add id in
                    if Fc_ast_state.is_type sym then
                       TokTypeId (sym, pos)
                    else
                       TokId (sym, pos)
      }

   (*
    * Some special symbols.
    *)
 | special
   { let pos = set_lexeme_position lexbuf in
        let s = Lexing.lexeme lexbuf in
           (* Some example keywords *)
           match s with
              ";" -> TokSemi pos
            | "+" -> TokPlus pos
            | "-" -> TokMinus pos
            | "(" -> TokLeftParen pos
            | ")" -> TokRightParen pos
            | _ -> raise (ParseError (pos, Printf.sprintf "illegal operator '%s'" (String.escaped s)))
   }

   (*
    * If all else fails, this is an error.
    *)
 | _
      { (* If nothing else, this is an error *)
        let pos = set_lexeme_position lexbuf in
        let s = Lexing.lexeme lexbuf in
           raise (ParseError (pos, Printf.sprintf "illegal character '%s'" (String.escaped s)))
      }

   (*
    * End of file.
    *)
 | eof
      { let pos = set_lexeme_position lexbuf in
           TokEof pos
      }
