(*
 * Right now the symbol table is just a representation of strings.
 *
 * ----------------------------------------------------------------
 *
 * Copyright (C) 1999 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * jyh@cs.caltech.edu
 *)

(*
 * We use a hashtable to manage symbols we have seen.
 *)
type symbol = int * string

type t =
   { hash : (string, symbol) Hashtbl.t;
     mutable count : int
   }

(*
 * A new symbol table.
 *)
let table =
   { hash = Hashtbl.create 19;
     count = 0
   }

(*
 * Add a symbol to the table.
 *)
let add s =
   try Hashtbl.find table.hash s with
      Not_found ->
         let index = succ table.count in
         let symbol = index, s in
            table.count <- index;
            Hashtbl.add table.hash s symbol;
            symbol

(*
 * Add a debugging symbol.
 *)
let debug_add s =
   try
      let len = String.length s in
      let i = String.rindex s '_' in
      let j = int_of_string (String.sub s (i + 1) (len - i - 1)) in
      let s = String.sub s 0 i in
         j, s
   with
      Not_found
    | Failure _ ->
         add s

(*
 * Get the symbol name.
 *)
let to_string (_, s) =
   s

(*
 * Create a new symbol.
 * Don't add it to the table.
 *)
let new_symbol_string v =
   let i = succ table.count in
      table.count <- i;
      i, v

let new_symbol (_, v) =
   new_symbol_string v

(*
 * Printer.
 * If the symbol is not a defined symbol,
 * print the index.
 *)
let print_symbol (i, s) =
   Format.print_string s;
   try
      let i', _ = Hashtbl.find table.hash s in
         if i' <> i then
            begin
               Format.print_string "_";
               Format.print_int i
            end
   with
      Not_found ->
         Format.print_string "_";
         Format.print_int i

(*
 * Compare for equality.
 *)
let eq (i1, _) (i2, _) =
   i1 = i2

let compare (i1, _) (i2, _) =
   i1 - i2

(*
 * Build a table where the key is a symbol.
 *)
module Base =
struct
   type elt = symbol
   let compare = compare
end

module SymbolSet = Red_black_set.MakeSet (Base)
module SymbolTable = Red_black_table.MakeTable (Base)

(*
 * -*-
 * Local Variables:
 * Caml-master: "set"
 * End:
 * -*-
 *)
